#!/usr/bin/env python3
# thoth-python
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Representation of source (index) for Python packages."""

import logging
import re
from functools import lru_cache
from urllib.parse import urlparse
from datetime import datetime

from typing import Optional, List, Union, Generator

import attr
import requests
from bs4 import BeautifulSoup
from packaging.version import Version
from packaging.version import LegacyVersion
from packaging.version import parse as parse_version

from .exceptions import NotFound
from .exceptions import InternalError
from .exceptions import VersionIdentifierError
from .configuration import config
from .artifact import Artifact

from thoth.common.helpers import parse_datetime

_LOGGER = logging.getLogger(__name__)
LEGACY_URLS = {"https://pypi.python.org/simple": "https://pypi.org/simple"}


def normalize_url(url: str) -> str:
    """We normalize url to remove legacy urls."""
    if url in LEGACY_URLS:
        return LEGACY_URLS[url]
    return url


@attr.s(frozen=True, slots=True)
class Source:
    """Representation of source (Python index) for Python packages."""

    url = attr.ib(type=str, converter=normalize_url)
    name = attr.ib(type=str)
    verify_ssl = attr.ib(type=bool, default=True)
    warehouse = attr.ib(type=bool)
    warehouse_api_url = attr.ib(default=None, type=Optional[str])

    _NORMALIZED_PACKAGE_NAME_RE = re.compile("[a-z-]+")

    @name.default
    def default_name(self):
        """Create a name for source based on url if not explicitly provided."""
        parsed_url = urlparse(self.url)
        name = parsed_url.netloc + parsed_url.path
        return re.sub(r"[^A-Za-z0-9]", "-", name)

    @warehouse.default
    def warehouse_default(self):
        """Check if the given url is registered in default warehouses."""
        return self.url in config.warehouses

    def get_api_url(self):
        """Construct URL to Warehouse instance."""
        if not self.warehouse:
            raise NotImplementedError("Cannot retrieve API url for non-warehouse repository")

        if self.warehouse_api_url:
            return self.warehouse_api_url

        return self.url[: -len("/simple")] + "/pypi"

    @classmethod
    def from_dict(cls, dict_: dict):
        """Parse source from its dictionary representation."""
        _LOGGER.debug("Parsing package source to dict representation")
        dict_ = dict(dict_)
        warehouse_url = dict_.pop("url")
        instance = cls(
            name=dict_.pop("name"),
            url=warehouse_url,
            verify_ssl=dict_.pop("verify_ssl"),
            warehouse=dict_.pop("warehouse", warehouse_url in config.warehouses),
        )

        if dict_:
            _LOGGER.warning("Ignored source configuration entries: %s", dict_)

        return instance

    def to_dict(self, include_warehouse: bool = False) -> dict:
        """Convert source definition to its dict representation."""
        _LOGGER.debug("Converting package source to dict representation")
        result = {"url": self.url, "verify_ssl": self.verify_ssl, "name": self.name}

        if include_warehouse:
            result["warehouse"] = self.warehouse

        return result

    @staticmethod
    def normalize_package_name(package_name: str) -> str:
        """Normalize package name on index according to PEP-0503."""
        return re.sub(r"[-_.]+", "-", package_name).lower()

    def _warehouse_get_api_package_version_info(self, package_name: str, package_version: str) -> dict:
        """Use API of the deployed Warehouse to gather package version information."""
        url = self.get_api_url() + f"/{package_name}/{package_version}/json"
        _LOGGER.debug("Gathering package version information from Warehouse API: %r", url)
        response = requests.get(url, verify=self.verify_ssl)
        if response.status_code == 404:
            raise NotFound(
                f"Package {package_name} in version {package_version} not found on warehouse {self.url} ({self.name})"
            )
        response.raise_for_status()
        return response.json()

    def _warehouse_get_api_package_info(self, package_name: str) -> dict:
        """Use API of the deployed Warehouse to gather package information."""
        url = self.get_api_url() + f"/{package_name}/json"
        _LOGGER.debug("Gathering package information from Warehouse API: %r", url)
        response = requests.get(url, verify=self.verify_ssl)
        if response.status_code == 404:
            raise NotFound(f"Package {package_name} not found on warehouse {self.url} ({self.name})")
        response.raise_for_status()
        return response.json()

    def _warehouse_get_package_hashes(
        self, package_name: str, package_version: str, with_included_files: bool = False
    ) -> List[dict]:
        """Gather information about SHA hashes available for the given package-version release."""
        package_info = self._warehouse_get_api_package_version_info(package_name, package_version)

        result = []
        for item in package_info["urls"]:
            result.append({"name": item["filename"], "sha256": item["digests"]["sha256"]})
            # this checks whether to gather digests for all files in the given artifact
            if with_included_files:
                artifact = Artifact(item["filename"], item["url"], verify_ssl=self.verify_ssl)
                result[-1]["digests"] = artifact.gather_hashes()
                result[-1]["symbols"] = artifact.get_versioned_symbols()

        return result

    @classmethod
    def is_normalized_python_package_name(cls, package_name: str) -> bool:
        """Check if the given Python package name is normalized."""
        # https://www.python.org/dev/peps/pep-0503/#normalized-names
        return cls._NORMALIZED_PACKAGE_NAME_RE.match(package_name) is not None

    @lru_cache(maxsize=10)
    def get_packages(self) -> set:
        """List packages available on the source package index."""
        _LOGGER.debug(f"Discovering packages available on {self.url} (simple index name: {self.name})")
        response = requests.get(self.url, verify=self.verify_ssl)
        response.raise_for_status()
        soup = BeautifulSoup(response.text, "lxml")
        links = soup.find_all("a")

        packages = set()
        for link in links:
            package_parts = link["href"].rsplit("/", maxsplit=2)
            # According to PEP-503, package names must have trailing '/', but check this explicitly
            if not package_parts[-1]:
                package_parts = package_parts[:-1]
            package_name = package_parts[-1]

            # Discard links to parent dirs (package name of URL does not match the text.
            link_text = link.text
            if link_text.endswith("/"):
                # Link names should end with trailing / according to PEEP:
                #   https://www.python.org/dev/peps/pep-0503/
                link_text = link_text[:-1]

            # Normalize link_text for comparison below
            link_text = self.normalize_package_name(link_text)

            if self.is_normalized_python_package_name(package_name) and package_name == link_text:
                packages.add(package_name)

        return packages

    def provides_package(self, package_name: str) -> bool:
        """Check if the given package is provided by this package source index."""
        _LOGGER.debug("Checking availability of package %r on index %r", package_name, self.url)
        package_name = self.normalize_package_name(package_name)
        url = self.url + "/" + package_name
        response = requests.get(url, verify=self.verify_ssl)

        if response.status_code == 404:
            return False

        # Raise on other inappropriate error codes.
        response.raise_for_status()
        return True

    @staticmethod
    def _parse_artifact_version(package_name: str, artifact_name: str) -> str:
        """Parse package version based on artifact name available on the package source index."""
        if artifact_name.endswith(".tar.gz"):
            # +1 for dash delimiting package name and package version.
            version = artifact_name[len(package_name) + 1 : -len(".tar.gz")]

        elif artifact_name.endswith(".whl"):
            # TODO: we will need to improve this based on PEP-0503.
            parsed_package_name, version, _ = artifact_name.split("-", maxsplit=2)
            if parsed_package_name.lower() != package_name:
                _LOGGER.warning(
                    f"It looks like package name does not match the one parsed from artifact when "
                    f"parsing version from wheel - package name is {package_name}, "
                    f"pared version is {version}, artifact is {artifact_name}"
                )
        else:
            raise InternalError(
                f"Unable to parse artifact name from artifact {artifact_name} for package {package_name}"
            )

        _LOGGER.debug(f"Parsed package version for package {package_name} from artifact {artifact_name}: {version}")
        return version

    def _simple_repository_list_versions(self, package_name: str) -> list:
        """List versions of package available on a simple repository."""
        simple_repos = set()
        for artifact_name, _ in self._simple_repository_list_artifacts(package_name):
            simple_repos.add(self._parse_artifact_version(package_name, artifact_name))

        result = list(simple_repos)
        _LOGGER.debug("Versions available on %r (index with name %r): %r", self.url, self.name, result)
        return result

    @lru_cache(maxsize=10)
    def get_package_versions(self, package_name: str) -> list:
        """Get listing of versions available for the given package."""
        if not self.warehouse:
            return self._simple_repository_list_versions(package_name)

        package_info = self._warehouse_get_api_package_info(package_name)
        return list(package_info["releases"].keys())

    def get_sorted_package_versions(
        self,
        package_name: str,
        graceful: bool = False,
        reverse: bool = True,  # default to newest first
    ) -> Optional[List[Union[Version, LegacyVersion]]]:
        """Get sorted versions for the given package."""
        try:
            all_versions = self.get_package_versions(package_name)
        except NotFound:
            if graceful:
                _LOGGER.warning(f"Package {package_name!r} was not found on index {self.name} ({self.url!r})")
                return None
            raise

        # Transform versions to semver for sorting:
        semver_versions = []
        for version in all_versions:
            try:
                version = parse_version(version)
            except Exception as exc:
                error_msg = f"Cannot parse semver version {version} for package {package_name}: {str(exc)}"
                if graceful:
                    _LOGGER.warning(error_msg)
                    return None
                raise VersionIdentifierError(error_msg) from exc

            semver_versions.append(version)
        return sorted(semver_versions, reverse=reverse)

    def get_latest_package_version(
        self, package_name: str, graceful: bool = False
    ) -> Optional[Union[Version, LegacyVersion]]:
        """Get the latest version for the given package."""
        semver_versions = self.get_sorted_package_versions(package_name=package_name, graceful=graceful)
        if semver_versions is None:  # only None if graceful so don't check
            return None
        return semver_versions[0]

    def _simple_repository_list_artifacts(self, package_name: str) -> list:
        """Parse simple repository package listing (HTML) and return artifacts present there."""
        url = self.url + "/" + package_name

        _LOGGER.debug(f"Discovering package {package_name} artifacts from {url}")
        response = requests.get(url, verify=self.verify_ssl)
        if response.status_code == 404:
            raise NotFound(f"Package {package_name} is not present on index {self.url} (index {self.name})")
        response.raise_for_status()
        soup = BeautifulSoup(response.text, "lxml")

        links = soup.find_all("a")
        artifacts = []
        for link in links:
            artifact_names = str(link["href"]).rsplit("/", maxsplit=1)
            if len(artifact_names) == 2:
                # If full URL provided by index.
                artifact_name = artifact_names[1]
            else:
                artifact_name = artifact_names[0]

            artifact_parts = artifact_name.rsplit("#", maxsplit=1)
            if len(artifact_parts) == 2:
                artifact_name = artifact_parts[0]

            if not artifact_name.endswith((".tar.gz", ".whl")):
                _LOGGER.debug("Link does not look like a package for %r: %r", package_name, link["href"])
                continue

            artifact_url = link["href"]
            if not artifact_url.startswith(("http://", "https://")):
                artifact_url = url + f"/{artifact_name}"

            artifacts.append((artifact_name, artifact_url))

        return artifacts

    def get_package_artifacts(self, package_name: str, package_version: str):
        """Return list of artifacts corresponding to package name and package version."""
        to_return = []
        possible_continuations = [".win32", ".tar.gz", ".whl", ".zip", ".exe", ".egg", "-"]
        for artifact_name, artifact_url in self._simple_repository_list_artifacts(package_name):
            # Convert all artifact names to lowercase - as a shortcut we simply convert everything to lowercase.
            artifact_name.lower()

            for i in possible_continuations:
                if artifact_name.startswith(f"{package_name}-{package_version}{i}"):
                    break
            else:
                _LOGGER.debug(
                    "Skipping artifact %r as it does not match required version %r for package %r",
                    artifact_name,
                    package_version,
                    package_name,
                )
                continue
            to_return.append(Artifact(artifact_name, artifact_url, verify_ssl=self.verify_ssl))

        return to_return

    def _download_artifacts_data(
        self, package_name: str, package_version: str, with_included_files: bool = False
    ) -> Generator[tuple, None, None]:
        """Download the given artifact from Warehouse and compute desired info."""
        possible_continuations = [".win32", ".tar.gz", ".whl", ".zip", ".exe", ".egg", "-"]
        for artifact_name, artifact_url in self._simple_repository_list_artifacts(package_name):
            # Convert all artifact names to lowercase - as a shortcut we simply convert everything to lowercase.
            artifact_name.lower()

            for i in possible_continuations:
                if artifact_name.startswith(f"{package_name}-{package_version}{i}"):
                    break
            else:
                _LOGGER.debug(
                    "Skipping artifact %r as it does not match required version %r for package %r",
                    artifact_name,
                    package_version,
                    package_name,
                )
                continue
            artifact = Artifact(artifact_name, artifact_url, verify_ssl=self.verify_ssl)

            symbols = None
            hashes = None
            if with_included_files:
                symbols = artifact.get_versioned_symbols()
                hashes = artifact.gather_hashes()

            yield (
                artifact_name,
                artifact.sha,
                hashes,
                symbols,
            )

    def provides_package_version(self, package_name: str, package_version: str) -> bool:
        """Check if the given source provides package in the given version."""
        try:
            return package_version in self.get_package_versions(package_name)
        except NotFound:
            # Package was not found on the index.
            return False

    @lru_cache(maxsize=10)
    def get_package_data(self, package_name: str, package_version: str, with_included_files: bool = False) -> list:
        """Get information about release hashes and symbols available in this source index."""
        if self.warehouse:
            return self._warehouse_get_package_hashes(package_name, package_version, with_included_files)

        artifacts = self.get_package_artifacts(package_name, package_version)
        result = []
        for artifact in artifacts:
            doc = {}
            doc["name"] = artifact.artifact_name
            doc["sha256"] = artifact.sha
            if with_included_files:
                doc["digests"] = artifact.gather_hashes()
                doc["symbols"] = artifact.get_versioned_symbols()
            result.append(doc)

        return result

    @lru_cache(maxsize=10)
    def get_package_hashes(self, package_name: str, package_version: str, with_included_files: bool = False) -> list:
        """Get information about release hashes available in this source index."""
        if self.warehouse:
            return self._warehouse_get_package_hashes(package_name, package_version, with_included_files)

        artifacts = self.get_package_artifacts(package_name, package_version)
        result = []
        for artifact in artifacts:
            doc = {}
            doc["name"] = artifact.artifact_name
            doc["sha256"] = artifact.sha
            if with_included_files:
                doc["digests"] = artifact.gather_hashes()
            result.append(doc)

        return result

    def get_package_release_date(
        self,
        package_name: str,
        package_version: str,
    ) -> datetime:
        """Get time at which package was uploaded to package index."""
        package_json = self._warehouse_get_api_package_info(package_name)
        release = package_json["releases"].get(package_version)
        if release is None:
            raise NotFound(f"Version {package_version} not found for {package_name} on {self.warehouse_api_url}.")
        artifact = next(x for x in release if x["python_version"] == "source")
        if artifact is None:
            raise NotFound(
                f"No source distribution for {package_name}==={package_version} found on {self.warehouse_api_url}."
            )

        return parse_datetime(artifact["upload_time_iso_8601"][:-1])



#!/usr/bin/env python3
# thoth-python
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Fetching package artifact digests."""

import typing
import logging

from .source import Source
from .exceptions import NotFound

_LOGGER = logging.getLogger(__name__)


class DigestsFetcherBase:
    """A base class implementing digests fetching."""

    def fetch_digests(self, package_name: str, package_version: str) -> dict:
        """Fetch digests for the given package in specified version from the given package index."""
        raise NotImplementedError


class PythonDigestsFetcher(DigestsFetcherBase):
    """Fetch digests from the given PEP-503 compatible package source index."""

    def __init__(self, sources: typing.List[Source]):
        """Set a list of package sources that should be considered when obtaining package digests."""
        self._sources = sources

    def fetch_digests(self, package_name: str, package_version: str) -> dict:
        """Fetch digests for the given package in specified version from the given package index."""
        report = {}

        for source in self._sources:
            try:
                report[source.url] = source.get_package_hashes(package_name, package_version)
            except NotFound as exc:
                _LOGGER.debug(
                    f"Package {package_name} in version {package_version} not "
                    f"found on index {source.name}: {str(exc)}"
                )

        return report



#!/usr/bin/env python3
# thoth-python
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Project abstraction and operations on project dependencies."""

import os
import logging
from itertools import chain
import tempfile

from typing import Optional, Dict, Any, List, Tuple

import attr
from thoth.common import cwd
from thoth.common import RuntimeEnvironment
from thoth.analyzer import run_command
from thoth.analyzer import CommandError

from .digests_fetcher import DigestsFetcherBase
from .digests_fetcher import PythonDigestsFetcher
from .pipfile import Pipfile
from .pipfile import PipfileLock
from .pipfile import PipfileMeta
from .source import Source
from .package_version import PackageVersion
from .exceptions import UnableLock
from .exceptions import InternalError
from .exceptions import NotFound
from .exceptions import FileLoadError
from .helpers import parse_requirements

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class Project:
    """A representation of a Python project."""

    pipfile = attr.ib(type=Pipfile)
    pipfile_lock = attr.ib(type=Optional[PipfileLock])
    runtime_environment = attr.ib(type=RuntimeEnvironment, default=attr.Factory(RuntimeEnvironment.from_dict))
    _graph_db = attr.ib(default=None)
    _workdir = attr.ib(default=None)

    @property
    def workdir(self) -> str:
        """Access working directory of project."""
        if self._workdir:
            return self._workdir

        self._workdir = tempfile.mkdtemp()
        return self._workdir

    @classmethod
    def from_files(
        cls,
        pipfile_path: Optional[str] = None,
        pipfile_lock_path: Optional[str] = None,
        *,
        runtime_environment: Optional[RuntimeEnvironment] = None,
        without_pipfile_lock: bool = False,
    ):
        """Create project from Pipfile and Pipfile.lock files."""
        try:
            pipfile = Pipfile.from_file(pipfile_path)
        except Exception as exc:
            raise FileLoadError(
                f"Failed to load Pipfile (path: {os.getcwd() if not pipfile_path else pipfile_path}: {str(exc)}"
            ) from exc

        pipfile_lock = None
        if not without_pipfile_lock:
            try:
                pipfile_lock = PipfileLock.from_file(pipfile_lock_path, pipfile=pipfile)
            except Exception as exc:
                raise FileLoadError(
                    f"Failed to load Pipfile.lock "
                    f"(path: {os.getcwd() if not pipfile_lock_path else pipfile_lock_path}: {str(exc)}"
                ) from exc

        return cls(
            pipfile,
            pipfile_lock,
            runtime_environment=runtime_environment if runtime_environment else RuntimeEnvironment.from_dict({}),
        )

    @classmethod
    def from_dict(
        cls,
        pipfile: Dict[str, Any],
        pipfile_lock: Dict[str, Any],
        runtime_environment: Optional[RuntimeEnvironment] = None,
    ) -> "Project":
        """Construct project out of a dict representation."""
        pip = Pipfile.from_dict(pipfile)
        if runtime_environment:
            return cls(
                pipfile=pip,
                pipfile_lock=PipfileLock.from_dict(pipfile_lock, pipfile=pip),
                runtime_environment=runtime_environment,
            )

        return cls(
            pipfile=pip,
            pipfile_lock=PipfileLock.from_dict(pipfile_lock, pipfile=pip),
        )

    @classmethod
    def from_strings(
        cls, pipfile_str: str, pipfile_lock_str: str = None, *, runtime_environment: RuntimeEnvironment = None
    ):
        """Create project from Pipfile and Pipfile.lock loaded into strings."""
        pipfile = Pipfile.from_string(pipfile_str)

        pipfile_lock = None
        if pipfile_lock_str:
            pipfile_lock = PipfileLock.from_string(pipfile_lock_str, pipfile)

        return cls(
            pipfile,
            pipfile_lock,
            runtime_environment=runtime_environment if runtime_environment else RuntimeEnvironment.from_dict({}),
        )

    def to_files(
        self,
        pipfile_path: str = None,
        pipfile_lock_path: str = None,
        *,
        without_pipfile_lock: bool = False,
        keep_thoth_section: bool = False,
    ) -> None:
        """Write the current state of project into Pipfile and Pipfile.lock files."""
        with open(pipfile_path or "Pipfile", "w") as pipfile_file:
            pipfile_file.write(self.pipfile.to_string(keep_thoth_section=keep_thoth_section))

        if not without_pipfile_lock:
            with open(pipfile_lock_path or "Pipfile.lock", "w") as pipfile_lock_file:
                pipfile_lock_file.write(self.pipfile_lock.to_string())  # type: ignore

    def construct_requirements_in(self) -> str:
        """Construct requirements.in file for the current project."""
        return self.pipfile.construct_requirements_in()

    def construct_requirements_txt(self) -> str:
        """Construct requirements.txt file for the current project - pip-tools compatible."""
        if self.pipfile_lock is not None:
            return self.pipfile_lock.construct_requirements_txt()
        else:
            raise TypeError("PipfileLock was not provided.")

    def to_pip_compile_files(
        self,
        requirements_path: str = "requirements.in",
        requirements_lock_path: str = "requirements.txt",
        without_lock: bool = False,
    ) -> None:
        """Write the current state of project into requirements.in and requirements.txt files.

        Files created are compatible with pip/pip-tools. If no lock is specified, requirements.txt
        hold unpinned direct dependencies as in case of pip.
        """
        if without_lock:
            with open(requirements_path, "w") as requirements_file:
                requirements_file.write(self.construct_requirements_in())
            return

        with open(requirements_path, "w") as requirements_file:
            requirements_file.write(self.construct_requirements_in())

        with open(requirements_lock_path, "w") as requirements_lock_file:
            requirements_lock_file.write(self.construct_requirements_txt())

    @classmethod
    def from_pip_compile_files(
        cls,
        requirements_path: str = "requirements.in",
        requirements_lock_path: Optional[str] = None,
        allow_without_lock: bool = False,
        runtime_environment: Optional[RuntimeEnvironment] = None,
    ) -> "Project":
        """Parse project from files compatible with pip/pip-tools."""
        sources_lock, package_versions_lock = None, None
        if allow_without_lock:
            if not os.path.exists(requirements_path):
                requirements_lock_path = requirements_lock_path or "requirements.txt"
                sources, package_versions = parse_requirements(requirements_lock_path)
            else:
                sources, package_versions = parse_requirements(requirements_path)
                if requirements_lock_path is not None:
                    sources_lock, package_versions_lock = parse_requirements(requirements_lock_path)
                elif os.path.exists("requirements.txt"):
                    sources_lock, package_versions_lock = parse_requirements("requirements.txt")
        else:
            requirements_lock_path = requirements_lock_path or "requirements.txt"
            sources, package_versions = parse_requirements(requirements_path)
            sources_lock, package_versions_lock = parse_requirements(requirements_lock_path)

        for package_version in package_versions_lock or []:
            if not package_version.is_locked():
                raise ValueError(
                    f"Package {package_version.name} in lockfile does not "
                    f"use locked version: {package_version.version!r}"
                )

            if not package_version.hashes:
                _LOGGER.warning(
                    "Parsed package %r in version %r from lockfile %r has no hashes assigned",
                    package_version.name,
                    package_version.version,
                    requirements_lock_path,
                )

        sources = sources or sources_lock or []

        # We know which index is assigned.
        if len(sources) == 1:
            for package_version in package_versions:
                package_version.index = sources[0]
            for package_version in package_versions_lock or []:
                package_version.index = sources[0]

        # XXX: do we want a default source?
        meta = PipfileMeta(sources={s.name: s for s in sources})
        pipfile = Pipfile.from_package_versions(package_versions, meta=meta)
        pipfile_lock = None
        if package_versions_lock:
            pipfile_lock = PipfileLock.from_package_versions(pipfile=pipfile, packages=package_versions_lock, meta=meta)

        if runtime_environment:
            return cls(pipfile, pipfile_lock, runtime_environment=runtime_environment)

        return cls(pipfile, pipfile_lock)

    @classmethod
    def from_package_versions(
        cls,
        packages: List[PackageVersion],
        packages_locked: List[PackageVersion] = None,
        meta: PipfileMeta = None,
        *,
        runtime_environment: RuntimeEnvironment = None,
    ):
        """Create project from PackageVersion objects.

        If locked packages are omitted, the lock has to be explicitly performed to generate
        in-memory Pipfile.lock representation.
        """
        pipfile = Pipfile.from_package_versions(packages, meta=meta)
        pipfile_lock = None
        if packages_locked:
            pipfile_lock = PipfileLock.from_package_versions(pipfile, packages_locked, meta=pipfile.meta)

        if runtime_environment:
            instance = cls(pipfile, pipfile_lock, runtime_environment=runtime_environment)
        else:
            # Let default be expanded.
            instance = cls(pipfile, pipfile_lock)

        instance.sanitize_source_indexes()
        return instance

    def set_allow_prereleases(self, allow_prereleases: bool = True) -> None:
        """Allow or disallow pre-releases for this project."""
        self.pipfile.meta.pipenv["allow_prereleases"] = allow_prereleases  # type: ignore

    @property
    def prereleases_allowed(self) -> bool:
        """Check if pre-releases are allowed for this project."""
        return self.pipfile.meta.pipenv.get("allow_prereleases", False) if self.pipfile.meta.pipenv else False

    def set_python_version(self, python_version: str = None) -> None:
        """Set version of Python used in the project."""
        if python_version is None:
            self.pipfile.meta.requires.pop("python_version")
        else:
            if self.pipfile.meta.requires is None:
                self.pipfile.meta.requires = {}

            self.pipfile.meta.requires["python_version"] = python_version

    @property
    def python_version(self) -> Optional[str]:
        """Get Python version used in this project."""
        if not self.pipfile.meta.requires:
            return None
        return self.pipfile.meta.requires.get("python_version", None)

    def to_dict(self):
        """Create a dictionary representation of this project."""
        return {
            "requirements": self.pipfile.to_dict(),
            "requirements_locked": self.pipfile_lock.to_dict() if self.pipfile_lock else None,
            "runtime_environment": self.runtime_environment.to_dict(),
        }

    def get_configuration_check_report(self) -> Optional[Tuple[Optional[dict], List[dict]]]:
        """Get a report on project configuration for the given runtime environment."""
        result = []
        changes_in_config = False

        # We check Python version if there is used Pipfile, it should match runtime configuration.
        pipfile_python_version = self.pipfile.meta.requires.get("python_version")
        runtime_python_version = self.runtime_environment.python_version

        # Keep the current runtime environment untouched.
        recommended_runtime_configuration_entry = RuntimeEnvironment.from_dict(self.runtime_environment.to_dict())

        if not pipfile_python_version and not runtime_python_version:
            result.append(
                {
                    "type": "WARNING",
                    "justification": "Please specify Python version in Pipfile using `pipenv --python <VERSION>` "
                    "and in Thoth's configuration file to have reproducible deployment and "
                    "recommendations targeting specific Python version",
                }
            )
        elif pipfile_python_version and not runtime_python_version:
            changes_in_config = True
            recommended_runtime_configuration_entry.python_version = pipfile_python_version
            result.append(
                {
                    "type": "WARNING",
                    "justification": "Use Python version in Thoth's configuration file to have "
                    "recommendations matching configuration in Pipfile, configured Python version "
                    f"in Pipfile is {pipfile_python_version}",
                }
            )

        # if pipfile_python_version and runtime_python_version:
        # This case is not related to adjustments in Thoth's configuration but rather in Pipfile - that is handled
        # in scoring.
        if not result:
            return None

        return recommended_runtime_configuration_entry.to_dict(without_none=True) if changes_in_config else None, result

    def get_outdated_package_versions(self, devel: bool = True) -> dict:
        """Get outdated packages in the lock file.

        This function will check indexes configured and look for version of
        each package. If the given package package is not latest, it will add
        it to the resulting list with the newer version identifier found on
        package index.
        """
        if not self.pipfile_lock:
            raise InternalError("Cannot check outdated packages on not-locked application stack")

        result = {}
        for package_version in self.iter_dependencies_locked(with_devel=devel):
            if package_version.index:
                latest = package_version.index.get_latest_package_version(package_version.name)
                if package_version.semantic_version != latest:
                    result[package_version.name] = (package_version, latest)
            else:
                found = False
                for index in self.pipfile_lock.meta.sources.values():
                    try:
                        latest = index.get_latest_package_version(package_version.name)
                        found = True
                    except NotFound:
                        continue

                    if package_version.semantic_version != latest:
                        result[package_version.name] = (package_version, latest)

                if not found:
                    raise NotFound(
                        f"Package {package_version!r} was not found on any package index"
                        f"configured: {self.pipfile_lock.meta.to_dict()}"
                    )

        return result

    def pipenv_lock(self):
        """Perform pipenv lock on the current state of project."""
        with cwd(self.workdir):
            self.pipfile.to_file()
            _LOGGER.debug("Running pipenv lock")

            try:
                result = run_command("pipenv lock", env={"PIPENV_IGNORE_VIRTUALENVS": "1"})
            except CommandError as exc:
                _LOGGER.exception(
                    "Unable to lock application stack (return code: %d):\n%s\n", exc.return_code, exc.stdout, exc.stderr
                )
                raise UnableLock("Failed to perform lock") from exc

            _LOGGER.debug("pipenv stdout:\n%s", result.stdout)
            _LOGGER.debug("pipenv stderr:\n%s", result.stderr)
            self.pipfile_lock = PipfileLock.from_file(pipfile=self.pipfile)

    def is_direct_dependency(self, package_version: PackageVersion) -> bool:
        """Check whether the given package is a direct dependency."""
        return package_version.name in self.pipfile.packages

    def get_locked_package_version(self, package_name: str) -> Optional[PackageVersion]:
        """Get locked version of the package."""
        if self.pipfile_lock is None:
            raise ValueError("PipfileLock was not provided and is None.")

        package_version = self.pipfile_lock.dev_packages.get(package_name)

        if not package_version:
            package_version = self.pipfile_lock.packages.get(package_name)

        return package_version

    def get_package_version(self, package_name: str) -> Optional[PackageVersion]:
        """Get locked version of the package."""
        package_version = self.pipfile.dev_packages.get(package_name)

        if not package_version:
            package_version = self.pipfile.packages.get(package_name)

        return package_version

    def exclude_package(self, package_version: PackageVersion) -> None:
        """Exclude the given package from application stack."""
        if not package_version.is_locked():
            raise InternalError("Cannot exclude package not pinned down to specific version: %r", package_version)

        section = self.pipfile.dev_packages if package_version.develop else self.pipfile.packages

        to_exclude_package = section.get(package_version.name)
        if to_exclude_package:
            package_version.negate_version()
            _LOGGER.debug(f"Excluding package {to_exclude_package} with package specified {package_version}")

            if package_version.index != to_exclude_package.index:
                _LOGGER.warning(
                    f"Excluding package {to_exclude_package} with {package_version} but package has "
                    f"different index configuration"
                )

            if package_version.markers != to_exclude_package:
                _LOGGER.warning(
                    f"Excluding package {to_exclude_package} with {package_version} but package has "
                    f"different markers configured"
                )

            # We do operations on the package we passed via parameters so the original package is
            # not adjusted if referenced elsewhere.
            if to_exclude_package.version != "*":
                package_version.version = f"{package_version.version},{to_exclude_package.version}"
            _LOGGER.debug("Package with excluded version %r", package_version)
            section[package_version.name] = package_version
        else:
            # Adding a new requirement.
            package_version.negate_version()
            section[package_version.name] = package_version
            _LOGGER.debug("Added excluded package to pipfile configuration: %r", package_version)

        self.pipenv_lock()

    def iter_dependencies_locked(self, with_devel: bool = True):
        """Iterate through locked dependencies of this project."""
        if not self.pipfile_lock:
            raise InternalError("Unable to chain locked dependencies - no Pipfile.lock provided")
        if with_devel:
            yield from chain(self.pipfile_lock.packages, self.pipfile_lock.dev_packages)
        else:
            yield from self.pipfile_lock.packages

    def iter_dependencies(self, with_devel: bool = True):
        """Iterate through direct dependencies of this project (not locked dependencies)."""
        if with_devel:
            yield from chain(self.pipfile.packages, self.pipfile.dev_packages)
        else:
            yield from self.pipfile.packages

    def _check_sources(self, whitelisted_sources: Optional[List]) -> List:
        """Check sources configuration in the Pipfile and report back any issues."""
        report = []
        for source in self.pipfile.meta.sources.values():
            if whitelisted_sources and source.url not in whitelisted_sources:
                report.append(
                    {
                        "type": "ERROR",
                        "id": "SOURCE-NOT-WHITELISTED",
                        "justification": f"Provided index {source.name!r} is not stated in the "
                        f"whitelisted package sources listing",
                        "source": source.to_dict(),
                    }
                )
            elif not source.verify_ssl:
                report.append(
                    {
                        "type": "WARNING",
                        "id": "INSECURE-SOURCE",
                        "justification": f"Source {source.name!r} does not use SSL/TLS verification",
                        "source": source.to_dict(),
                    }
                )

        return report

    def _index_scan(self, digests_fetcher: DigestsFetcherBase) -> Tuple[list, dict]:
        """Generate full report for packages given the sources configured."""
        report = {}
        findings = []
        if self.pipfile_lock is None:
            raise ValueError("PipfileLock was not provided and has value None.")

        for package_version in chain(self.pipfile_lock.packages, self.pipfile_lock.dev_packages):
            if package_version.name in report:
                # TODO: can we have the same package in dev packages and packages?
                _LOGGER.warning(f"Package {package_version.name} already present in the report")
                continue

            index_report = digests_fetcher.fetch_digests(package_version.name, package_version.locked_version)
            findings.extend(self._check_scan(package_version, index_report))
            report[package_version.name] = index_report

        return findings, report

    def _check_scan(self, package_version: PackageVersion, index_report: dict) -> list:
        """Find and report errors found in the index scan."""
        scan_report = []

        # Prepare hashes for inspection.
        hashes = []
        pipenv_style_hashes = []
        for entry in index_report.values():
            hashes.append([item["sha256"] for item in entry])
            pipenv_style_hashes.append([f"sha256:{item['sha256']}" for item in entry])

        if not package_version.index and len(index_report.keys()) > 1:
            # Are there indexes with different artifacts present? - suggest to specify index explicitly.
            if set(chain(*hashes)) != set(hashes[0]):
                sources_reported = ", ".join(index_report.keys())
                scan_report.append(
                    {
                        "type": "WARNING",
                        "id": "DIFFERENT-ARTIFACTS-ON-SOURCES",
                        "justification": f"Configured sources ({sources_reported}) have different artifacts "
                        f"available, specify explicitly source to be used",
                        "indexes": list(index_report.keys()),
                        "package_locked": package_version.to_pipfile_lock(),
                        "package_name": package_version.name,
                        "package_version": package_version.version,
                        "sources": index_report,
                    }
                )

        # Source for reports.
        source = None
        if package_version.index:
            source = package_version.index.to_dict()  # type: ignore

        if package_version.index and index_report.get(package_version.index.url) and len(hashes) > 1:
            # Is installed from different source - which one?
            used_package_version_hashes = set(h[len("sha256:") :] for h in package_version.hashes)  # type: ignore
            configured_index_hashes = set(h["sha256"] for h in index_report[package_version.index.url])

            # Find other sources from which artifacts can be installed.
            other_sources = {}  # type: Dict[str, List[Tuple]]
            for artifact_hash in package_version.hashes:
                artifact_hash = artifact_hash[  # type: ignore
                    len("sha256:") :
                ]  # Remove pipenv-specific hash formatting.  # type: ignore

                for index_url, index_info in index_report.items():
                    if index_url == package_version.index.url:
                        # Skip index that is assigned to the package, we are inspecting from the other sources.
                        continue

                    artifact_entry = [(i.get("name"), i["sha256"]) for i in index_info if i["sha256"] == artifact_hash]

                    if not artifact_entry:
                        continue

                    if index_url not in other_sources:
                        other_sources[index_url] = []

                    other_sources[index_url].extend(artifact_entry)

            if not set(used_package_version_hashes).issubset(set(configured_index_hashes)):
                # Source is different from the configured one.
                scan_report.append(
                    {
                        "type": "ERROR",
                        "id": "ARTIFACT-DIFFERENT-SOURCE",
                        "justification": f"Artifacts are installed from different "
                        f'sources ({",".join(other_sources.keys())}) not respecting configuration',
                        "source": source,  # type: ignore
                        "package_locked": package_version.to_pipfile_lock(),
                        "package_name": package_version.name,
                        "package_version": package_version.version,
                        "indexes": list(other_sources.keys()),
                        "sources": other_sources,
                    }
                )
            elif other_sources:
                # Warn about possible installation from another source.
                scan_report.append(
                    {
                        "type": "INFO",
                        "id": "ARTIFACT-POSSIBLE-DIFFERENT-SOURCE",
                        "justification": f"Artifacts can be installed from different sources "
                        f'({",".join(other_sources.keys())}) not respecting configuration '
                        f"that expects {package_version.index.name!r}",
                        "source": source,  # type: ignore
                        "package_locked": package_version.to_pipfile_lock(),
                        "package_name": package_version.name,
                        "package_version": package_version.version,
                        "indexes": list(other_sources.keys()),
                        "sources": other_sources,
                    }
                )

        if package_version.index and not index_report.get(package_version.index.url):
            # Configured index does not provide the given package.
            scan_report.append(
                {
                    "type": "ERROR",
                    "id": "MISSING-PACKAGE",
                    "justification": f"Source index {package_version.index.name!r} explicitly "
                    f"assigned to package {package_version.name!r} but there was not found "
                    f"any record for the given package",
                    "source": source,  # type: ignore
                    "package_locked": package_version.to_pipfile_lock(),
                    "package_name": package_version.name,
                    "package_version": package_version.version,
                    "sources": index_report,
                }
            )

        # Changed hashes?
        for digest in package_version.hashes:
            digest = digest[len("sha256:") :]  # type: ignore
            for index_info in index_report.values():
                if any(item["sha256"] == digest for item in index_info):
                    break
            else:
                scan_report.append(
                    {
                        "type": "ERROR",
                        "id": "INVALID-ARTIFACT-HASH",
                        "justification": "Hash for the given artifact was not found",
                        "source": source,  # type: ignore
                        "package_locked": package_version.to_pipfile_lock(),
                        "package_name": package_version.name,
                        "package_version": package_version.version,
                        "digest": digest,  # type: ignore
                    }
                )

        return scan_report

    def check_provenance(self, whitelisted_sources: list = None, digests_fetcher: DigestsFetcherBase = None) -> List:
        """Check provenance/origin of packages that are stated in the project."""
        digests_fetcher = digests_fetcher or PythonDigestsFetcher(list(self.pipfile.meta.sources.values()))
        findings, _ = self._index_scan(digests_fetcher)
        findings.extend(self._check_sources(whitelisted_sources))
        return findings

    def sanitize_source_indexes(self):
        """Make sure all the indexes are correctly propagated to Pipfile and Pipfile.lock metadata."""
        self.pipfile.sanitize_source_indexes()
        if self.pipfile_lock:
            self.pipfile_lock.sanitize_source_indexes()

    def add_source(
        self,
        url: str,
        verify_ssl: bool = True,
        name: str = None,
        warehouse: bool = False,
        warehouse_api_url: str = None,
    ) -> Source:
        """Add a package source (index) to the project."""
        if name:
            source = Source(
                name=name, url=url, verify_ssl=verify_ssl, warehouse=warehouse, warehouse_api_url=warehouse_api_url
            )
        else:
            # Do not provide source index name so that attrs correctly compute default based on URL.
            source = Source(url=url, verify_ssl=verify_ssl, warehouse=warehouse, warehouse_api_url=warehouse_api_url)

        self.pipfile.meta.add_source(source)
        if self.pipfile_lock:
            self.pipfile_lock.meta.add_source(source)

        return source

    def add_package(
        self, package_name: str, package_version: str = None, *, source: Source = None, develop: bool = False
    ):
        """Add the given package to project.

        This method will add packages only to Pipfile, locking has to be explicitly done once package is added.
        """
        if source and source.name not in self.pipfile.meta.sources:
            raise InternalError(
                f"Adding package {package_name} to project without having source index "
                f"{source.name} registered in the project"
            )

        pkg_vers = PackageVersion(
            name=package_name, version=package_version, develop=develop, index=source if source else None
        )
        self.pipfile.add_package_version(pkg_vers)



#!/usr/bin/env python3
# thoth-python
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Representation of development and default packages as stated in Pipfile and Pipfile.lock."""

import typing
import logging

import attr

from .package_version import PackageVersion
from .exceptions import InternalError
from .exceptions import PackageVersionAlreadyPresentError

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class Packages:
    """Encapsulate logic on package manipulation."""

    develop = attr.ib(type=bool)
    packages = attr.ib(type=dict)

    def is_develop(self):
        """Check if packages provided by this instance are development packages."""
        return self.develop

    def is_default(self):
        """Check if packages provided by this instance are dependencies of application packages."""
        return not self.develop

    def to_pipfile(self) -> dict:
        """Convert packages representation as seen in Pipfile file."""
        _LOGGER.debug("Generating Pipfile entry for packages (develop: %s)", self.develop)
        result = {}
        for package_name, package_version in self.packages.items():
            result.update(package_version.to_pipfile())

        return result

    def to_pipfile_lock(self) -> dict:
        """Convert packages representation as seen in Pipfile.lock file."""
        _LOGGER.debug("Generating Pipfile.lock entry for packages (develop: %s)", self.develop)
        result = {}
        for package_name, package_version in self.packages.items():
            result.update(package_version.to_pipfile_lock())

        return result

    @classmethod
    def from_package_versions(cls, package_versions: typing.List[PackageVersion], develop: bool):
        """Create Packages instance from a list of packages in specific versions."""
        if not package_versions:
            return cls(develop=develop, packages={})

        package_version_map = {}
        for package_version in package_versions:
            if package_version.name in package_version_map:
                raise InternalError(f"Atempt adding multiple packages with same name to Packages: {package_version!r}")

            package_version_map[package_version.name] = package_version
            if develop != package_version.develop:
                raise InternalError(
                    "Not all packages provided to construct Packages instance have the same develop flag set"
                )

        return cls(develop=develop, packages=package_version_map)

    @classmethod
    def from_pipfile(cls, packages, develop, meta):
        """Parse Pipfile entry stating list of packages used."""
        _LOGGER.debug("Parsing Pipfile entry for %s packages", "develop" if develop else "default")
        package_version = {}
        for package_name, package_info in packages.items():
            package_version[package_name] = PackageVersion.from_pipfile_entry(package_name, package_info, develop, meta)

        return cls(develop=develop, packages=package_version)

    @classmethod
    def from_pipfile_lock(cls, packages, develop, meta):
        """Parse Pipfile.lock entry stating list of packages used."""
        _LOGGER.debug("Parsing Pipfile.lock entry for %s packages", "develop" if develop else "default")
        package_version = {}
        for package_name, package_info in packages.items():
            package_version[package_name] = PackageVersion.from_pipfile_lock_entry(
                package_name, package_info, develop, meta
            )

        return cls(develop=develop, packages=package_version)

    def __iter__(self):
        """Iterate over packages encapsulated by this wrapper."""
        yield from self.packages.values()

    def get(self, package_name: str) -> typing.Optional[PackageVersion]:
        """Get package by its name."""
        return self.packages.get(package_name)

    def __setitem__(self, package_name: str, package_version: PackageVersion) -> None:
        """Set the given package to a value."""
        self.packages[package_name] = package_version

    def __getitem__(self, item):
        """Get the given package from section."""
        return self.packages[item]

    def add_package_version(self, package_version: PackageVersion, *, force: bool = False):
        """Add the given package version to package list."""
        if (package_version.develop and not self.develop) or (not package_version.develop and self.develop):
            raise InternalError(f"Adding package {package_version!r} to package listing without proper develop flag")

        if package_version.name in self.packages and not force:
            raise PackageVersionAlreadyPresentError(
                f"Adding package {package_version!r} to packages, but this package is already present there"
            )

        self.packages[package_version.name] = package_version



#!/usr/bin/env python3
# thoth-python
# Copyright(C) 2019, 2020 Kevin Postlthwait
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Representation a python module and all the files within."""

import shutil
import logging
import requests
import tempfile
import zipfile
import tarfile
import hashlib
import os
from elftools.elf.elffile import ELFFile
from typing import Iterator, Tuple, Any
import attr

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class Artifact:
    """Python artifacts are compressed modules."""

    artifact_name = attr.ib(type=str)
    artifact_url = attr.ib(type=str)
    compressed_file = attr.ib(type=str, default=None)
    dir_name = attr.ib(type=str, default=None)
    verify_ssl = attr.ib(type=bool, default=False)
    sha = attr.ib(type=str, default=None)

    def __attrs_post_init__(self):
        """Initialize sha after init."""
        self.sha = self._calculate_sha()

    def _download_if_necessary(self):
        if self.compressed_file is None:
            self._download_artifact()

    def _extract_if_necessary(self):
        self._download_if_necessary()
        if self.dir_name is None:
            self._extract_py_module()

    def _download_artifact(self) -> None:
        _LOGGER.debug("Downloading artifact from url %r", self.artifact_url)
        response = requests.get(self.artifact_url, verify=self.verify_ssl, stream=True)
        response.raise_for_status()
        with tempfile.NamedTemporaryFile(mode="w+b", delete=False) as f:
            self.compressed_file = f.name
            f.write(response.content)

    def _extract_py_module(self) -> None:

        self._download_if_necessary()

        try:
            self.dir_name = tempfile.mkdtemp()
            try:
                if self.compressed_file.endswith(".tar.gz"):
                    tf = tarfile.open(self.compressed_file)
                    tf.extractall(self.dir_name)
                    _LOGGER.debug("Artifact is .tar.gz file")
                else:
                    ext = self.compressed_file.split(".")[-1]
                    with zipfile.ZipFile(self.compressed_file) as zip_ref:
                        zip_ref.extractall(self.dir_name)
                        _LOGGER.debug("Artifact is .%r file", ext)
            except Exception as e:
                _LOGGER.exception(f"Could not extract {self.compressed_file}: {str(e)}")
        except Exception as exc:
            _LOGGER.exception(f"Could not create temp dir: {str(exc)}")

    def _calculate_sha(self) -> str:
        """Calculate SHA256 of compressed file if not present in url."""
        url_parts = self.artifact_url.rsplit("#", maxsplit=1)
        if len(url_parts) == 2 and url_parts[1].startswith("sha256="):
            sha256 = url_parts[1][len("sha256=") :]
            _LOGGER.debug("Using SHA256 stated in URL: %r", url_parts[1])
            return sha256

        self._download_if_necessary()

        with open(self.compressed_file, "rb") as f:
            digest = hashlib.sha256()
            chunk = 1024
            while True:
                data = f.read(chunk)
                if data:
                    digest.update(data)
                else:
                    break

        hex_digest = digest.hexdigest()
        _LOGGER.debug("Computed artifact sha256 digest for %r: %s", self.artifact_url, digest)
        return hex_digest

    #                       VERSIONED SYMBOLS                                #
    def _elf_find_versioned_symbols(self, elf: ELFFile) -> Iterator[Tuple[str, str]]:
        """Take an ELFFile object and outputs the required dynamic symbols."""
        section = elf.get_section_by_name(".gnu.version_r")

        if section is not None:
            for verneed, verneed_iter in section.iter_versions():
                if verneed.name.startswith("ld-linux"):
                    continue
                for vernaux in verneed_iter:
                    yield verneed.name, vernaux.name

    def _is_elf(self, filename: str) -> bool:
        """Check if files magic numbers are <delete>ELF."""
        with open(filename, "rb") as f:
            return f.read(16).startswith(b"\x7f\x45\x4c\x46")

    def _get_versioned_symbols_from_file(self, result, filename: str):
        """Given a file get all required dynamic symbols if it's an executable."""
        with open(filename, "rb") as f:
            if not self._is_elf(filename):
                return
            elf = ELFFile(f)
            for lib, sym in self._elf_find_versioned_symbols(elf):
                if result.get(lib) is None:
                    result[lib] = set()
                result[lib].add(sym)

    def get_versioned_symbols(self) -> dict:
        """Walk dir and get all dynamic symbols required from all files."""
        self._extract_if_necessary()
        result = dict()  # type: Any
        for dir_name, _, file_list in os.walk(self.dir_name):
            for fname in file_list:
                self._get_versioned_symbols_from_file(result, os.path.join(dir_name, fname))

        for symbol in result:
            result[symbol] = list(result[symbol])

        return result

    #                          Package Digests                                          #
    def gather_hashes(self) -> list:
        """Calculate checksums and gather hashes of all file in the given artifact."""
        self._extract_if_necessary()

        digests = []
        for root, _, files in os.walk(self.dir_name):
            for file_ in files:
                filepath = os.path.join(root, file_)
                if os.path.isfile(filepath):
                    with open(filepath, "rb") as my_file:
                        digests.append(
                            {
                                "filepath": filepath[len(self.dir_name) + 1 :],
                                "sha256": hashlib.sha256(my_file.read()).hexdigest(),
                            }
                        )
        return digests

    def __del__(self):
        """Remove temporary file created by class."""
        try:
            if self.compressed_file.startswith("/tmp"):
                os.remove(self.compressed_file)
            shutil.rmtree(self.dir_name)
        except Exception:
            pass



#!/usr/bin/env python3
# thoth-python
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Helper functions and utilities."""

from typing import Tuple
from typing import List
from typing import Dict
from typing import Set
from typing import Any
from typing import TYPE_CHECKING
from itertools import chain
import logging
from packaging.requirements import Requirement
from packaging.markers import Variable
from packaging.markers import Op
from packaging.markers import Value
from packaging.utils import canonicalize_name

from .source import Source
from .package_version import PackageVersion
from .exceptions import FileLoadError

if TYPE_CHECKING:
    from .project import Project

_LOGGER = logging.getLogger(__name__)


def fill_package_digests(generated_project: "Project") -> "Project":
    """Temporary fill package digests stated in Pipfile.lock."""
    if generated_project.pipfile_lock is None:
        raise ValueError("Generated project PipfileLock is not set.")
    for package_version in chain(generated_project.pipfile_lock.packages, generated_project.pipfile_lock.dev_packages):
        if package_version.hashes:
            # Already filled from the last run.
            continue

        if package_version.index:
            scanned_hashes = package_version.index.get_package_hashes(
                package_version.name, package_version.locked_version
            )
        else:
            for source in generated_project.pipfile.meta.sources.values():
                try:
                    scanned_hashes = source.get_package_hashes(package_version.name, package_version.locked_version)
                    break
                except Exception:
                    continue
            else:
                raise ValueError("Unable to find package hashes")

        for entry in scanned_hashes:
            package_version.hashes.append("sha256:" + entry["sha256"])

    return generated_project


def _marker_reduction(marker, extra):  # type: ignore
    """Convert internal packaging marker representation to interpretation which can be evaluated.

    As markers also depend on `extra' which will cause issues when evaluating marker in the solver
    environment, let's substitute `extra' marker with a condition which evaluates always to true.
    """
    if isinstance(marker, str):
        return marker

    if isinstance(marker, list):
        result_markers = []
        for nested_marker in marker:
            reduced_marker = _marker_reduction(nested_marker, extra)  # type: ignore
            result_markers.append(reduced_marker)

        return result_markers

    if marker[0].value != "extra":
        return marker

    extra.add(str(marker[2]))
    # A special case to handle extras in markers - substitute extra with a marker which always evaluates to true:
    return Variable("python_version"), Op(">="), Value("0.0")


def parse_requirement_str(requirement_str: str) -> Dict[str, Any]:
    """Parse a string representation of marker."""
    # Some notes on this implementation can be found at: https://github.com/pypa/packaging/issues/211
    requirement = Requirement(requirement_str)

    evaluation_result = None
    evaluation_error = None
    extra = set()  # type: Set[str]
    marker_evaluated_str = None
    marker_str = str(requirement.marker) if requirement.marker else None
    if requirement.marker:
        # We perform a copy of marker specification during the traversal so that we
        # do not evaluate "extra" marker - according to PEP-508, this behavior
        # raises an error if the interpreting environment does not explicitly
        # define them. As we are aggregating "generic" data and extra is
        # user-defined on the actual resolution, we exclude this extra marker
        # here.
        markers_copy = []
        for marker in requirement.marker._markers:
            marker_copy = _marker_reduction(marker, extra)  # type: ignore
            markers_copy.append(marker_copy)

        try:
            requirement.marker._markers = markers_copy
            evaluation_result = requirement.marker.evaluate()
            marker_evaluated_str = str(requirement.marker)
        except Exception as exc:
            _LOGGER.exception("Failed to evaluate marker {}".format(requirement.marker))
            evaluation_error = str(exc)
    else:
        evaluation_result = True

    return {
        "package_name": requirement.name,
        "normalized_package_name": canonicalize_name(requirement.name),
        "specifier": str(requirement.specifier) if requirement.specifier else None,
        "resolved_versions": [],
        "extras": list(requirement.extras),
        "extra": list(extra),
        "marker": marker_str,
        "marker_evaluated": marker_evaluated_str,
        "marker_evaluation_result": evaluation_result,
        "marker_evaluation_error": evaluation_error,
    }


def parse_requirements(file_path: str) -> Tuple[List[Source], List[PackageVersion]]:
    """Parse requirements.{txt,in} file."""
    sources = []
    package_versions = []

    try:
        with open(file_path, "r") as input_file:
            content = input_file.read()
    except Exception as exc:
        raise FileLoadError(f"Failed to load requirements file at {file_path!r}: {str(exc)}") from exc

    # Remove escaped new lines.
    content = content.replace("\\\n", "")
    for line in content.splitlines():
        line = line.lstrip()
        if line.startswith("#"):
            continue

        if not line:
            continue

        if line.startswith(("-i", "--index-url", "--extra-index-url")):
            index_url = line.split(" ", maxsplit=1)[1]
            sources.append(Source(index_url))
            continue

        if line.startswith("-"):
            _LOGGER.warning("Ignoring line in %r: %s", file_path, line)
            continue

        parts = line.split("--hash=")
        hashes = parts[1:]

        req = parse_requirement_str(parts[0])
        package_versions.append(
            PackageVersion(
                name=req["normalized_package_name"],
                version=req["specifier"] or "*",
                extras=req["extras"],
                markers=req["marker"] or None,
                hashes=hashes,
                develop=False,  # XXX: we could add Thoth's specific comment for develop packages.
            )
        )

    return sources, package_versions



#!/usr/bin/env python3
# thoth-python
# Copyright(C) 2018, 2019, 2020 Red Hat, Inc.
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Python ecosystem specific routines."""

from .digests_fetcher import DigestsFetcherBase
from .digests_fetcher import PythonDigestsFetcher
from .packages import Packages
from .package_version import PackageVersion
from .pipfile import Pipfile
from .pipfile import PipfileMeta
from .pipfile import PipfileLock
from .project import Project
from .source import Source
from .aiosource import AIOSource, AsyncIterablePackages, AsyncIterableVersions, AsyncIterableArtifacts


__version__ = "0.15.4"
__author__ = "Fridolin Pokorny <fridolin@redhat.com>, Christoph Görn <goern@redhat.com>"
__copyright__ = "Copyright 2018, 2019 Red Hat, Inc."
__license__ = "GPLv3+"


__all__ = [
    "DigestsFetcherBase",
    "PythonDigestsFetcher",
    "Packages",
    "PackageVersion",
    "Pipfile",
    "PipfileMeta",
    "PipfileLock",
    "Project",
    "Source",
    "AIOSource",
    "AsyncIterablePackages",
    "AsyncIterableVersions",
    "AsyncIterableArtifacts",
]



#!/usr/bin/env python3
# thoth-python
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Representation of source (index) for Python packages."""

import logging

from functools import lru_cache

import attr
import asyncio
import aiohttp

from bs4 import BeautifulSoup

from .exceptions import NotFound
from .artifact import Artifact
from .source import Source

from typing import Optional, Set, Tuple, List, Dict

_LOGGER = logging.getLogger(__name__)


class AsyncIterablePackages:
    """Async Iterator for Packages."""

    def __init__(self, _packages: Set[str]):  # Ignore PyDocStyleBear
        """Initialize Async Iterable instance for Packages."""
        self.packages = _packages

    def __aiter__(self):  # Ignore PyDocStyleBear
        """Return asynchronous iterator."""
        return self

    async def __anext__(self) -> str:  # Ignore PyDocStyleBear
        """Return awaitable object."""
        data = await self.fetch_data()

        if data:
            return data
        else:
            raise StopAsyncIteration

    async def fetch_data(self) -> Optional[str]:  # Ignore PyDocStyleBear
        """Fetch data."""
        await asyncio.sleep(0.1)  # Other coros get to run

        if len(self.packages) == 0:
            return None

        return self.packages.pop()


class AsyncIterableVersions:
    """Async Iterator for Versions."""

    def __init__(self, _versions: Set[str]):  # Ignore PyDocStyleBear
        """Initialize Async Iterable Versions instance for Versions."""
        self.versions = _versions

    def __aiter__(self):  # Ignore PyDocStyleBear
        """Return asynchronous iterator."""
        return self

    async def __anext__(self) -> str:  # Ignore PyDocStyleBear
        """Return awaitable object."""
        data = await self.fetch_data()

        if data:
            return data
        else:
            raise StopAsyncIteration

    async def fetch_data(self) -> Optional[str]:  # Ignore PyDocStyleBear
        """Fetch data."""
        await asyncio.sleep(0.1)  # Other coros get to run

        if len(self.versions) == 0:
            return None

        return self.versions.pop()


class AsyncIterableArtifacts:
    """Async Iterator for Artifacts."""

    def __init__(self, artifacts: List[Tuple[str, str]]):  # Ignore PyDocStyleBear
        """Initialize Async Iterable instance for Artifacts."""
        self.artifacts = artifacts

    def __aiter__(self):  # Ignore PyDocStyleBear
        """Return asynchronous iterator."""
        return self

    async def __anext__(self) -> Tuple:  # Ignore PyDocStyleBear
        """Return awaitable object."""
        data = await self.fetch_data()

        if data:
            return data
        else:
            raise StopAsyncIteration

    async def fetch_data(self) -> Optional[Tuple]:  # Ignore PyDocStyleBear
        """Fetch data."""
        await asyncio.sleep(0.1)  # Other coros get to run

        if len(self.artifacts) == 0:
            return None

        return self.artifacts.pop()


@attr.s(frozen=True, slots=True)
class AIOSource(Source):
    """Representation of source (Python index) for Python packages."""

    async def _warehouse_get_api_package_version_info(  # type: ignore
        self, package_name: str, package_version: str
    ) -> Dict:
        """Use API of the deployed Warehouse to gather package version information."""
        url = self.get_api_url() + f"/{package_name}/{package_version}/json"

        _LOGGER.debug("Gathering package version information from Warehouse API: %r", url)

        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.get(url) as response:
                if response.status == 404:
                    raise NotFound(
                        f"Package {package_name} in version {package_version} not "
                        f"found on warehouse {self.url} ({self.name})"
                    )

                return await response.json()

    async def _warehouse_get_package_hashes(  # type: ignore
        self, package_name: str, package_version: str, with_included_files: bool = False
    ) -> List[Dict]:
        """Gather information about SHA hashes available for the given package-version release."""
        package_info = await self._warehouse_get_api_package_version_info(package_name, package_version)

        result = []
        for item in package_info["urls"]:
            result.append({"name": item["filename"], "sha256": item["digests"]["sha256"]})
            # this checks whether to gather digests for all files in the given artifact
            if with_included_files:
                artifact = Artifact(item["filename"], item["url"], verify_ssl=self.verify_ssl)
                result[-1]["digests"] = artifact.gather_hashes()
                result[-1]["symbols"] = artifact.get_versioned_symbols()

        return result

    async def _warehouse_get_api_package_info(self, package_name: str) -> Dict:  # type: ignore
        """Use API of the deployed Warehouse to gather package information."""
        url = self.get_api_url() + f"/{package_name}/json"

        _LOGGER.debug("Gathering package information from Warehouse API: %r", url)

        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.get(url) as response:
                if response.status == 404:
                    raise NotFound(f"Package {package_name} not found on warehouse {self.url} ({self.name})")

                return await response.json()

    async def _simple_repository_list_versions(self, package_name: str) -> List:  # type: ignore
        """List versions of package available on a simple repository."""
        result = set()
        a = await self._simple_repository_list_artifacts(package_name)

        async for artifact_name, _ in a:
            result.add(self._parse_artifact_version(package_name, artifact_name))

        to_ret = list(result)

        _LOGGER.debug("Versions available on %r (index with name %r): %r", self.url, self.name, result)
        return to_ret

    async def _simple_repository_list_artifacts(self, package_name: str) -> AsyncIterableArtifacts:  # type: ignore
        """Parse simple repository package listing (HTML) and return artifacts present there."""
        url = self.url + "/" + package_name
        links = []

        _LOGGER.debug("Discovering package %r artifacts from %r", package_name, url)

        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.get(url) as response:
                if response.status == 404:
                    raise NotFound(f"Package {package_name} is not present on index {self.url} (index {self.name})")

                text = await response.text()
                soup = BeautifulSoup(text, "lxml")

                links = soup.find_all("a")

        artifacts = []  # List[Tuple[str,str]]
        for link in links:
            artifact_name_full = str(link["href"]).rsplit("/", maxsplit=1)
            if len(artifact_name_full) == 2:
                # If full URL provided by index.
                artifact_name = artifact_name_full[1]
            else:
                artifact_name = artifact_name_full[0]

            artifact_parts = artifact_name.rsplit("#", maxsplit=1)
            if len(artifact_parts) == 2:
                artifact_name = artifact_parts[0]

            if not artifact_name.endswith((".tar.gz", ".whl")):
                _LOGGER.debug("Link does not look like a package for %r: %r", package_name, link["href"])
                continue

            artifact_url = link["href"]
            if not artifact_url.startswith(("http://", "https://")):
                artifact_url = url + f"/{artifact_name}"

            artifacts.append((artifact_name, artifact_url))

        return AsyncIterableArtifacts(artifacts)

    async def get_packages(self) -> Optional[AsyncIterablePackages]:  # type: ignore
        """List packages available on the source package index."""
        _LOGGER.debug(f"Discovering packages available on {self.url} (simple index name: {self.name})")

        links = []

        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.get(self.url) as resp:
                text = await resp.text()

                soup = BeautifulSoup(text, "lxml")
                links = soup.find_all("a")

        if len(links) == 0:
            return None

        packages = set()
        for link in links:
            package_parts = link["href"].rsplit("/", maxsplit=2)
            # According to PEP-503, package names must have trailing '/', but check this explicitly
            if not package_parts[-1]:
                package_parts = package_parts[:-1]
            package_name = self.normalize_package_name(package_parts[-1])

            # Discard links to parent dirs (package name of URL does not match the text.
            link_text = self.normalize_package_name(link.text)
            if link_text.endswith("/"):
                # Link names should end with trailing / according to PEEP:
                #   https://www.python.org/dev/peps/pep-0503/
                link_text = link_text[:-1]

            if package_name == link_text:
                packages.add(package_name)

        return AsyncIterablePackages(packages)

    @lru_cache(maxsize=64)
    async def get_package_versions(self, package_name: str) -> AsyncIterableVersions:  # type: ignore
        """Get listing of versions available for the given package."""
        if not self.warehouse:
            return AsyncIterableVersions(set(await self._simple_repository_list_versions(package_name)))

        package_info = await self._warehouse_get_api_package_info(package_name)
        return AsyncIterableVersions(set(package_info["releases"].keys()))

    async def get_package_artifacts(self, package_name: str, package_version: str):
        """Return list of artifacts corresponding to package name and package version."""
        to_return = []
        possible_continuations = [".win32", ".tar.gz", ".whl", ".zip", ".exe", ".egg", "-"]
        async for artifact_name, artifact_url in await self._simple_repository_list_artifacts(package_name):
            # Convert all artifact names to lowercase - as a shortcut we simply convert everything to lowercase.
            artifact_name.lower()

            for i in possible_continuations:
                if artifact_name.startswith(f"{package_name}-{package_version}{i}"):
                    break
            else:
                _LOGGER.debug(
                    "Skipping artifact %r as it does not match required version %r for package %r",
                    artifact_name,
                    package_version,
                    package_name,
                )
                continue
            to_return.append(Artifact(artifact_name, artifact_url, verify_ssl=self.verify_ssl))

        return to_return

    @lru_cache(maxsize=10)
    async def get_package_hashes(  # type: ignore
        self, package_name: str, package_version: str, with_included_files: bool = False
    ) -> List:
        """Get information about release hashes available in this source index."""
        if self.warehouse:
            return await self._warehouse_get_package_hashes(package_name, package_version, with_included_files)

        artifacts = await self.get_package_artifacts(package_name, package_version)
        result = []
        for artifact in artifacts:
            doc = {}
            doc["name"] = artifact.artifact_name
            doc["sha256"] = artifact.sha
            if with_included_files:
                doc["digests"] = artifact.gather_hashes()
            result.append(doc)

        return result



#!/usr/bin/env python3
# thoth-python
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Exceptions used in the thoth-python package."""


class ThothPythonException(Exception):
    """A base class for thoth-python exceptions."""


class DirectDependencyRemoval(ThothPythonException):
    """Raised if trying to remove direct dependency from application stack.

    Or there is no option to remove the given dependency from application stack.
    """


class UnableLock(ThothPythonException):
    """Raised if trying to lock invalid application stack or resolution cannot be done."""


class PipfileParseError(ThothPythonException):
    """An exception raised on invalid Pipfile or Pipfile.lock."""


class InternalError(ThothPythonException):
    """An exception raised on bugs in the code base."""


class PackageVersionAlreadyPresentError(ThothPythonException):
    """An exception raised when adding a package in specific version that is already present."""


class SourceNotFoundError(ThothPythonException):
    """An exception raise when the given package source is not found."""


class VersionIdentifierError(ThothPythonException):
    """An exception raised if the given version identifier is not a semver identifier."""


class UnsupportedConfiguration(ThothPythonException):
    """Raised on unsupported configuration options."""


class NotFound(ThothPythonException):
    """Raised if the given artifact cannot be found."""


class FileLoadError(ThothPythonException):
    """Raised when failed to open or parse a file."""



#!/usr/bin/env python3
# thoth-python
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Configuration used for computing recommendations."""

import os

import attr

_DEFAULT_WAREHOUSES = ("https://pypi.python.org/simple", "https://pypi.org/simple")


@attr.s(slots=True)
class _Configuration:
    """Configuration representation for recommendation engine.

    Accepted environment variables:
      * THOTH_PYTHON_WAREHOUSES - a JSON describing warehouses
    """

    warehouses = attr.ib(type=list)

    @warehouses.default
    def warehouses_default(self):
        warehouses = _DEFAULT_WAREHOUSES
        if "THOTH_PYTHON_WAREHOUSES" in os.environ:
            warehouses = os.environ["THOTH_PYTHON_WAREHOUSES"].split(",")

        return warehouses


config = _Configuration()



#!/usr/bin/env python3
# thoth-python
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Representation of packages in the application stack."""

import re
import logging
from copy import copy

import attr
from packaging.version import LegacyVersion
from packaging.version import parse as parse_version
from packaging.utils import canonicalize_name


from .exceptions import UnsupportedConfiguration
from .exceptions import PipfileParseError
from .exceptions import InternalError
from .source import Source

from typing import Optional, Tuple, Union
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .pipfile import PipfileMeta

_LOGGER = logging.getLogger(__name__)


_RE_NORMALIZE_PYTHON_PACKAGE_NAME = re.compile(r"[-_.]+")


def _normalize_python_package_name(package_name: str) -> str:
    """Normalize Python package name based on PEP-0503."""
    # Make sure we have normalized names in the graph database according to PEP:
    #   https://www.python.org/dev/peps/pep-0503/#normalized-names
    return canonicalize_name(package_name)


def _normalize_python_package_version(package_version: Optional[str]) -> Optional[str]:
    """Normalize Python package version based on PEP-440."""
    if package_version is None or package_version == "*":
        return "*"
    return str(parse_version(package_version))


class Version:
    """A simple wrapper around packaging's version to support seamless API for legacy and current version handling."""

    __slots__ = ["_version"]

    def __init__(self, version_identifier: str):
        """Initialize version."""
        self._version = parse_version(version_identifier)

    def __repr__(self) -> str:
        """Get version representation."""
        return repr(self._version)

    def __str__(self):
        """Get version string."""
        return str(self._version)

    def __lt__(self, other: "Version") -> bool:
        """Compare two versions."""
        return self._version.__lt__(other._version)

    def __le__(self, other: "Version") -> bool:
        """Compare two versions."""
        return self._version.__le__(other._version)

    def __eq__(self, other: object) -> bool:
        """Compare two versions."""
        if not isinstance(other, Version):
            raise NotImplementedError

        return self._version.__eq__(other._version)

    def __ge__(self, other: "Version") -> bool:
        """Compare two versions."""
        return self._version.__ge__(other._version)

    def __gt__(self, other: "Version") -> bool:
        """Compare two versions."""
        return self._version.__gt__(other._version)

    def __ne__(self, other: object) -> bool:
        """Compare two versions."""
        if not isinstance(other, Version):
            raise NotImplementedError

        return self._version.__ne__(other._version)

    @property
    def epoch(self) -> int:
        """Get version epoch."""
        return self._version.epoch

    @property
    def release(self) -> Tuple[int, ...]:
        """Get version release."""
        # LegacyVersion from packaging can return None here. That is not something we want to have. Return
        # an empty tuple instead.
        return self._version.release or tuple()

    @property
    def pre(self) -> Optional[Tuple[str, int]]:
        """Get version pre."""
        return self._version.pre

    @property
    def post(self) -> Optional[Tuple[str, int]]:
        """Get version post."""
        return self._version.post

    @property
    def dev(self) -> Optional[Tuple[str, int]]:
        """Get version dev."""
        return self._version.dev

    @property
    def local(self) -> Optional[str]:
        """Get version local."""
        return self._version.local

    @property
    def public(self) -> str:
        """Get version public."""
        return self._version.public

    @property
    def base_version(self) -> str:
        """Get version base."""
        return self._version.base_version

    @property
    def is_prerelease(self) -> bool:
        """Check if version is a pre-release."""
        return self._version.is_prerelease

    @property
    def is_postrelease(self) -> bool:
        """Check if version is a post-release."""
        return self._version.is_postrelease

    @property
    def is_devrelease(self) -> bool:
        """Check if version is a dev-release."""
        return self._version.is_devrelease

    @property
    def is_legacy_version(self) -> bool:
        """Check if the given version is a legacy version identifier."""
        return isinstance(self._version, LegacyVersion)

    @property
    def major(self) -> int:
        """Get version major release."""
        if isinstance(self._version, LegacyVersion):
            return 0  # Compatibility handling.
        return self._version.major

    @property
    def minor(self) -> int:
        """Get version minor release."""
        if isinstance(self._version, LegacyVersion):
            return 0  # Compatibility handling.
        return self._version.minor

    @property
    def micro(self) -> int:
        """Get version micro release."""
        if isinstance(self._version, LegacyVersion):
            return 0  # Compatibility handling.
        return self._version.micro


@attr.s(slots=True)
class PackageVersion:
    """A package version as described in the Pipfile.lock entry."""

    name = attr.ib(type=str, converter=_normalize_python_package_name)
    version = attr.ib(type=str, converter=_normalize_python_package_version)
    develop = attr.ib(type=bool)
    index = attr.ib(default=None, type=Optional[Source])
    hashes = attr.ib(default=attr.Factory(list))
    markers = attr.ib(default=None, type=Optional[str])
    extras = attr.ib(default=attr.Factory(list))
    _semantic_version = attr.ib(default=None, type=Version)
    _locked_version = attr.ib(default=None, type=Optional[str])
    _package_tuple = attr.ib(default=None, type=Optional[Tuple[str, str, Optional[str]]])
    _package_tuple_locked = attr.ib(default=None, type=Optional[Tuple[str, str, Optional[str]]])

    def to_dict(self) -> dict:
        """Create a dictionary representation of parameters (useful for later constructor calls)."""
        return {
            "name": self.name,
            "version": self.version,
            "develop": self.develop,
            "index": self.index,
            "hashes": self.hashes,
            "markers": self.markers,
            "extras": self.extras,
        }

    def __eq__(self, other):
        """Check for package-version equality."""
        return self.name == other.name and self.version == other.version and self.index.url == other.index.url

    def __lt__(self, other):
        """Compare same packages based on their semantic version."""
        if self.name != other.name:
            raise ValueError(f"Comparing package versions of different package - {self.name} and {other.name}")

        return self.semantic_version < other.semantic_version

    def __gt__(self, other):
        """Compare same packages based on their semantic version."""
        if self.name != other.name:
            raise ValueError(f"Comparing package versions of different package - {self.name} and {other.name}")

        return self.semantic_version > other.semantic_version

    @classmethod
    def normalize_python_package_name(cls, package_name: str) -> str:
        """Normalize Python package name based on PEP-0503.

        https://www.python.org/dev/peps/pep-0503/#normalized-names
        """
        return _normalize_python_package_name(package_name)

    @classmethod
    def normalize_python_package_version(cls, package_version: str) -> str:
        """Normalize Python package version based on PEP-440.

        https://www.python.org/dev/peps/pep-0440/#normalization
        """
        return _normalize_python_package_version(package_version)  # type: ignore

    @classmethod
    def from_model(cls, model, *, develop: bool = False):
        """Convert database model representation to object representation."""
        # TODO: add hashes to the graph database
        # TODO: we will need to add index information - later on?
        return cls(
            name=model.package_name, version=model.package_version, develop=develop, index=Source(url=model.index)
        )

    def is_locked(self):
        """Check if the given package is locked to a specific version."""
        return self.version.startswith("==")

    def duplicate(self):
        """Duplicate the given package safely when performing changes in resolution."""
        return PackageVersion(
            name=self.name,
            version=copy(self.version),
            develop=self.develop,
            index=self.index,
            hashes=self.hashes,
            markers=self.markers,
            extras=self.extras,
        )

    def negate_version(self) -> None:
        """Negate version of a locked package version."""
        if not self.is_locked() or self.version is None:
            raise InternalError(
                f"Negating version on non-locked package {self.name} with version {self.version} is not supported"
            )

        self.version = "!" + self.version[1:]

    @property
    def locked_version(self) -> str:
        """Retrieve locked version of the package."""
        if not self._locked_version:
            if not self.is_locked() or self.version is None:
                raise InternalError(
                    f"Requested locked version for {self.name} but package has no locked version {self.version}"
                )

            self._locked_version = self.version[len("==") :]

        return self._locked_version

    @property
    def semantic_version(self) -> Version:
        """Get semantic version respecting version specified - package has to be locked to a specific version."""
        if not self._semantic_version:
            if not self.is_locked():
                raise InternalError(
                    f"Cannot get semantic version for not-locked package {self.name} in version {self.version}"
                )

            self._semantic_version = Version(self.locked_version)

        return self._semantic_version

    @staticmethod
    def parse_semantic_version(version_identifier: str) -> Version:
        """Parse the given version identifier into a semver representation."""
        return Version(version_identifier)

    @staticmethod
    def _get_index_from_meta(meta: "PipfileMeta", package_name: str, index_name: Optional[str]) -> Optional[Source]:
        """Get the only index name present in the Pipfile.lock metadata.

        If there is no index explicitly assigned to package, there is only one package source
        index configured in the meta. Assign it to package.
        """
        if index_name is not None and index_name in meta.sources:
            return meta.sources[index_name]
        elif index_name is not None and index_name not in meta.sources:
            raise PipfileParseError(f"Configured index {index_name} for package {package_name} not found in metadata")
        # We could also do this branch, but that can be dangerous as SHAs might differ in Pipfile.lock.
        #
        # elif index_name is None and len(meta.sources) == 1:
        #    return list(meta.sources.values())[0]

        # Unfortunatelly Pipenv does not explicitly assign indexes to
        # packages, give up here with unassigned index
        return None

    @classmethod
    def from_pipfile_lock_entry(cls, package_name: str, entry: dict, develop: bool, meta: "PipfileMeta"):
        """Construct PackageVersion instance from representation as stated in Pipfile.lock."""
        _LOGGER.debug("Parsing entry in Pipfile.lock for package %r: %s", package_name, entry)
        entry = dict(entry)

        if any(not entry.get(conf) for conf in ("version", "hashes")):
            raise PipfileParseError(
                f"Package {package_name} has missing or empty configuration in the locked entry: {entry}"
            )

        instance = cls(
            name=package_name,
            version=entry.pop("version"),
            index=cls._get_index_from_meta(meta, package_name, entry.pop("index", None)),
            hashes=entry.pop("hashes"),
            markers=entry.pop("markers", None),
            develop=develop,
            extras=entry.pop("extras", []),
        )

        if entry:
            _LOGGER.warning(f"Unused entries when parsing Pipfile.lock for package {package_name}: {entry}")

        return instance

    def to_pipfile_lock(self) -> dict:
        """Create an entry as stored in the Pipfile.lock."""
        _LOGGER.debug("Generating Pipfile.lock entry for package %r", self.name)

        if not self.is_locked():
            raise InternalError(f"Trying to generate Pipfile.lock with packages not correctly locked: {self}")

        # TODO: uncomment once we will have hashes available in the graph
        # if not self.hashes:
        #     raise InternalError(f"Trying to generate Pipfile.lock without assigned hashes for package: {self}")

        result = {"version": self.version, "hashes": self.hashes}

        if self.markers:
            result["markers"] = self.markers

        if self.index:
            result["index"] = self.index.name

        if self.extras:
            result["extras"] = self.extras

        return {self.name: result}

    def to_tuple(self) -> Tuple[str, str, Optional[str]]:
        """Return a tuple representing this Python package."""
        if not self._package_tuple:
            self._package_tuple = self.name, self.locked_version, self.index.url if self.index else None

        return self._package_tuple

    def to_tuple_locked(self) -> Tuple[str, str, Optional[str]]:
        """Return a tuple representing this Python package - used for locked packages."""
        if not self._package_tuple_locked:
            self._package_tuple_locked = self.name, self.locked_version, self.index.url if self.index else None

        return self._package_tuple_locked

    def to_pipfile(self):
        """Generate Pipfile entry for the given package."""
        _LOGGER.debug("Generating Pipfile entry for package %r", self.name)
        result = {}

        if self.index:
            result["index"] = self.index.name

        if self.markers:
            result["markers"] = self.markers

        if self.extras:
            result["extras"] = self.extras

        if not result:
            # Only version information is available.
            return {self.name: self.version if self.version is not None else "*"}

        result["version"] = self.version if self.version is not None else "*"
        return {self.name: result}

    @classmethod
    def from_pipfile_entry(cls, package_name: str, entry: Union[dict, str], develop: bool, meta: "PipfileMeta"):
        """Construct PackageVersion instance from representation as stated in Pipfile."""
        _LOGGER.debug("Parsing entry in Pipfile for package %r: %s", package_name, entry)
        # Pipfile holds string for a version:
        #   thoth-storages = "1.0.0"
        # Or a dictionary with additional configuration:
        #   thoth-storages = {"version": "1.0.0", "index": "pypi"}
        index = None
        extras = []
        markers = None
        if isinstance(entry, str):
            package_version = entry
        else:
            if any(vcs in entry for vcs in ("git", "hg", "bzr", "svn")):
                raise UnsupportedConfiguration(
                    f"Package {package_name!r} uses a version control system instead of package index: {entry}"
                )

            if "editable" in entry:
                raise UnsupportedConfiguration(
                    f"Package {package_name!r} is editable local project instead of a package from a package index"
                )

            if "version" not in entry:
                raise UnsupportedConfiguration(
                    f"Package {package_name!r} does not state any version range specifier: {entry}"
                )

            package_version = entry.pop("version")
            index = entry.pop("index", None)
            extras = entry.pop("extras", [])
            markers = entry.pop("markers", None)

            if entry:
                _LOGGER.warning("Unparsed part of Pipfile: %s", entry)

        instance = cls(
            name=package_name,
            version=package_version,
            index=cls._get_index_from_meta(meta, package_name, index),
            develop=develop,
            extras=extras,
            markers=markers,
        )

        return instance



#!/usr/bin/env python3

"""Setup for python module."""

import os
import sys
from setuptools import setup
from pathlib import Path
from setuptools.command.test import test as test_command


def get_requirements():
    """Get requirements."""
    with open("requirements.txt") as fd:
        return fd.read().splitlines()


def get_version():
    """Get version."""
    with open(os.path.join("thoth", "python", "__init__.py")) as f:
        content = f.readlines()

    for line in content:
        if line.startswith("__version__ ="):
            # dirty, remove trailing and leading chars
            return line.split(" = ")[1][1:-2]

    raise ValueError("No package version found")


class Test(test_command):
    """Introduce test command to run testsuite using pytest."""

    _IMPLICIT_PYTEST_ARGS = [
        "tests/",
        "--timeout=600",
        "--cov=./thoth",
        "--mypy",
        "--capture=no",
        "--verbose",
        "-l",
        "-s",
        "-vv",
    ]

    user_options = [("pytest-args=", "a", "Arguments to pass into py.test")]

    def initialize_options(self):
        """Initialize options."""
        super().initialize_options()
        self.pytest_args = None

    def finalize_options(self):
        """Finalize options."""
        super().finalize_options()
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        """Run tests."""
        import pytest

        passed_args = list(self._IMPLICIT_PYTEST_ARGS)

        if self.pytest_args:
            self.pytest_args = [arg for arg in self.pytest_args.split() if arg]
            passed_args.extend(self.pytest_args)

        sys.exit(pytest.main(passed_args))


VERSION = get_version()

setup(
    name="thoth-python",
    version=VERSION,
    packages=["thoth.python"],
    package_data={"thoth.python": ["py.typed"]},
    install_requires=get_requirements(),
    author="Fridolin Pokorny",
    author_email="fridolin@redhat.com",
    maintainer="Fridolin Pokorny",
    maintainer_email="fridolin@redhat.com",
    description="A Python ecosystem specific library",
    long_description=Path("README.rst").read_text(),
    url="https://github.com/thoth-station/python",
    license="GPLv3+",
    keywords="python dependency pypi dependencies tool library thoth",
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: Implementation :: CPython",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    ],
    cmdclass={"test": Test},
    command_options={
        "build_sphinx": {
            "version": ("setup.py", VERSION),
            "release": ("setup.py", VERSION),
        }
    },
)
